;;; binary-data.scm

;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Internal Functions ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;

(define string-copy
  (lambda (str)
    (let* ((len (string-length str))
	   (new (make-string len)))
      (do ((i 0 (+ i 1)))
	  ((= i len) new)
	(string-set! new i (string-ref str i))))))

(define string-reverse
  (lambda (str)
    (let ((len (string-length str))
	  (new (string-copy str)))
      (do ((left 0 (+ left 1))
	   (right (- len 1) (- right 1)))
	  ((negative? right) new)
	(string-set! new left (string-ref str right))))))

(define string-reverse!
  (lambda (str)
    (let ((len (string-length str)))
      (if (> len 0)
	  (do ((left 0 (+ left 1))
	       (right (- len 1) (- right 1)))
	      ((>= left right) str)
	    (let ((tmp (string-ref str left)))
	      (string-set! str left (string-ref str right))
	      (string-set! str right tmp)))
	  ""))))


;;; (reverse-endian binary-data) => binary-data

(define reverse-endian
  (lambda (bd)
    (if (not (string? bd))
	(error "input not binary-data" bd)
	(string-reverse bd))))

(define reverse-endian!
  (lambda (bd)
    (if (not (string? bd))
	(error "input not binary-data" bd)
	(string-reverse! bd))))

(define big-endian->little-endian reverse-endian!)
(define little-endian->big-endian reverse-endian!)


;;; (twos-complement number)

(define twos-complement
  (lambda (num)
    (+ 1 (bitwise-not num))))


;;; (psab str) => unspecified
;;; print str as binary

(define psab
  (lambda (str)
    (for-each (lambda (char)
		(display (number->string (low-uint8->number (string char)) 2))
		(display #\space))
	      (string->list str))
    (newline)))


;;; (optional-sym list default)  => symbol
;;; looks at something like '(foo) and returns foo or default if null.

(define optional-sym
  (lambda (list default)
    (if (null? list)
	default
	(car list))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Scheme Number To Binary-Data ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; convert a scheme number into a string
;;; that contains the number as binary data
;;; calculations assume big endian, ie. msb followed by lsb

;;;
;;; unsigned
;;;

(define ref8  (expt 2 8))
(define ref16 (expt 2 16))
(define ref32 (expt 2 32))

(define max8  (- ref8 1))
(define max16 (- ref16 1))
(define max32 (- ref32 1))

(define range8-error  (lambda (n) (error (format #f "number out of range: 0 < n < ~a" ref8) n)))
(define range16-error (lambda (n) (error (format #f "number out of range: 0 < n < ~a" ref16) n)))
(define range32-error (lambda (n) (error (format #f "number out of range: 0 < n < ~a" ref32) n)))


;;; At the entry point, we create a string the size of the final value.
;;; We divide the number into 8 bit pieces and pass the string with
;;; the character offset to a function that fills that hole.

(define number->uint8/char ascii->char)

(define fill-hole!
  (lambda (str offset n8)
    (cond ((not (string? str))
	   (error "not a string" str))
	  ((not (integer? offset))
	   (error "not an integer" offset))
	  ((> n8 max8)
	   (error "number too big" n8))
	  ((negative? offset)
	   (error "offset too small" offset))
	  ((>= offset (string-length str))
	   (error "offset too big" offset))
	  (else (string-set! str offset (number->uint8/char n8))))))


;;; (number->uint8 number . end)  => number-as-binary-data

(define number->uint8/fill!
  (lambda (str offset n)
    (string-set! str offset (number->uint8/char n))))

(define number->uint8
  (lambda (n . end)
    (if (<= 0 n max8)
	(string (number->uint8/char n))
	(range8-error n))))


;;; (number->uint16 number . end)  => number-as-binary-data

(define number->uint16/fill!
  (lambda (str offset n . end)
    (let ((end (optional-sym end 'little-endian))
	  (msb (quotient n ref8))
	  (lsb (bitwise-and n max8))
	  (offset0 offset)
	  (offset1 (+ offset 1)))
      (cond ((not (<= 0 n max16)) (range16-error n))
	    ((eq? end 'big-endian)
	     (fill-hole! str offset0 msb)
	     (fill-hole! str offset1 lsb))
	    (else
	     (fill-hole! str offset1 msb)
	     (fill-hole! str offset0 lsb))))))

(define number->uint16
  (lambda (n . end)
    (let ((end (optional-sym end 'little-endian)))
      (let ((str (make-string 2)))
	(number->uint16/fill! str 0 n end)
	str))))


;;; (number->uint32 number . end)  => number-as-binary-data

(define number->uint32/fill!
  (let ((byte-size 4))
    (lambda (str offset n . end)
      (let ((end (optional-sym end 'little-endian)))
	(cond ((not (<= 0 n max32)) (range32-error n))
	      ((eq? end 'big-endian)
	       (do ((n n (arithmetic-shift n -8))
		    (byte (+ offset (- byte-size 1)) (- byte 1)))
		   ((< byte offset) #t)
		 (fill-hole! str byte (bitwise-and n #xff))))
	      (else
	       (do ((n n (arithmetic-shift n -8))
		    (byte offset (+ byte 1)))
		   ((>= byte (+ offset byte-size)) #t)
		 (fill-hole! str byte (bitwise-and n #xff)))))))))

(define number->uint32
  (let ((byte-size 4))
    (lambda (n . end)
      (let ((end (optional-sym end 'little-endian)))
	(let ((str (make-string 4)))
	  (number->uint32/fill! str 0 n end)
	  str)))))


;;;
;;; signed
;;;

(define sref8  (quotient (expt 2 8) 2))
(define sref16 (quotient (expt 2 16) 2))
(define sref32 (quotient (expt 2 32) 2))

(define smax8  (- sref8 1))
(define smax16 (- sref16 1))
(define smax32 (- sref32 1))

(define smin8  (- sref8))
(define smin16 (- sref16))
(define smin32 (- sref32))

(define ranges8-error
  (lambda (num)
    (error (format #f "number out of range: ~a < num < ~a" smin8 smax8) num)))
(define ranges16-error
  (lambda (num)
    (error (format #f "number out of range: ~a < num < ~a" smin16 smax16) num)))
(define ranges32-error
  (lambda (num)
    (error (format #f "number out of range: ~a < num < ~a" smin32 smax32) num)))


;;; (number->sint8 scheme-number)  => binary-data

(define number->sint8/fill!
  (lambda (str offset num)
    (cond ((<= 0 num smax8)  (number->uint8/fill! str offset num))
	  ((<= smin8 num -1)
	   (number->uint8/fill! str offset (bitwise-xor #xff (twos-complement num))))
	  (else (ranges8-error num)))))

(define number->sint8
  (lambda (num)
    (let ((str (make-string 1)))
      (number->sint8/fill! str 0 num)
      str)))


;;; (number->sint16 scheme-number)  => binary-data

(define number->sint16/fill!
  (lambda (str offset num . end)
    (let ((end (optional-sym end 'little-endian)))
      (cond ((<= 0 num smax16)
	     (number->uint16/fill! str offset num end))
	    ((<= smin16 num -1)
	     (number->uint16/fill! str offset (bitwise-xor #xffff (twos-complement num)) end))
	    (else (ranges16-error num))))))

(define number->sint16
  (lambda (num . end)
    (let ((end (optional-sym end 'little-endian))
	  (str (make-string 2)))
      (number->sint16/fill! str 0 num)
      str)))


;;; (number->sint32 scheme-number)  => binary-data

(define number->sint32/fill!
  (lambda (str offset num . end)
    (let ((end (optional-sym end 'little-endian)))
      (cond ((<= 0 num smax32)
	     (number->uint32/fill! str offset num end))
	    ((<= smin32 num -1)
	     (number->uint32/fill! str offset
				   (bitwise-xor #xffffffff (twos-complement num)) end))
	    (else (ranges32-error num))))))

(define number->sint32
  (lambda (num . end)
    (let ((end (optional-sym end 'little-endian))
	  (str (make-string 4)))
      (number->sint32/fill! str 0 num)
      str)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Binary Data to Scheme Number ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Take a string of binary data and convert
;;; it to a scheme number
;;; calculations assume big endian, ie. msb followed by lsb
;;; functions beginning with low- take strings as arguments
;;; while their unadorned counterparts take binary-data as arguments.

;;;
;;; unsigned
;;;

;;; (low-uint8->number binary-data-string)  => number

(define low-uint8->number
  (lambda (str)
    (char->ascii (string-ref str 0))))


;;; (uint8->number binary-data)  => number

(define uint8->number
  (lambda (bd . end)
    (if (not (string? bd))
	(error "input not binary-data" bd)
	(low-uint8->number bd))))


;;; (low-uint16->number binary-data-string)  => number
;;; always assumes big endian

(define low-uint16->number
  (lambda (str)
    (let ((msb (low-uint8->number (substring str 0 1)))
	  (lsb (low-uint8->number (substring str 1 2))))
      (+ (* msb 256) lsb))))


;;; (uint16->number binary-data . endian)  => number

(define uint16->number
  (lambda (bd . end)
    (if (not (string? bd))
	(error "input not binary-data" bd)
	(let ((end (optional-sym end 'little-endian)))
	  (if (eq? end 'big-endian)
	      (low-uint16->number bd)
	      (low-uint16->number (little-endian->big-endian bd)))))))


;;; (low-uint32->number binary-data-string)  => number
;;; always assumes big endian

(define low-uint32->number
  (lambda (str)
    (let ((ms2b (low-uint16->number (substring str 0 2)))
	  (ls2b (low-uint16->number (substring str 2 4))))
      (+ (* ms2b 65536) ls2b))))


;;; (uint32->number binary-data . endian)  => number

(define uint32->number
  (lambda (bd . end)
    (if (not (string? bd))
	(error "input not binary-data" bd)
	(let ((end (optional-sym end 'little-endian)))
	  (if (eq? end 'big-endian)
	      (low-uint32->number bd)
	      (low-uint32->number (little-endian->big-endian bd)))))))

;;;
;;; signed
;;;

;;; (low-sint8->number binary-data-string)  => scheme-number

(define low-sint8->number
  (lambda (str)
    (let* ((num (low-uint8->number str)))
      (if (< num 128)
	  num
	  (twos-complement (bitwise-xor #xff num))))))


;;; (sint8->number binary-data)  => scheme-number

(define sint8->number
  (lambda (bd . end)
    (if (not (string? bd))
	(error "input not binary-data" bd)
	(low-sint8->number bd))))


;;; (low-sint16->number binary-data-string end)  => scheme-number
;;; assumes big endian

(define low-sint16->number
  (lambda (str)
    (let* ((num (low-uint16->number str)))
      (if (< num 32768)
	  num
	  (twos-complement (bitwise-xor #xffff num))))))

;;; (sint16->number binary-data . end)  => scheme-number

(define sint16->number
  (lambda (bd . end)
    (if (not (string? bd))
	(error "input not binary-data" bd)
	(let ((end (optional-sym end 'little-endian)))
	  (if (eq? end 'big-endian)
	      (low-sint16->number bd)
	      (low-sint16->number (big-endian->little-endian bd)))))))


;;; (low-sint32->number binary-data-string end)  => scheme-number
;;; assumes big endian

(define low-sint32->number
  (let ((max (expt 2 31)))
    (lambda (str)
      (let* ((num (low-uint32->number str)))
	(if (< num max)
	    num
	    (twos-complement (bitwise-xor #xffffffff num)))))))


;;; (sint32->number binary-data . end)  => scheme-number

(define sint32->number
  (lambda (bd . end)
    (if (not (string? bd))
	(error "input not binary-data" bd)
	(let ((end (optional-sym end 'little-endian)))
	  (if (eq? end 'big-endian)
	      (low-sint32->number bd)
	      (low-sint32->number (big-endian->little-endian bd)))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Read Binary Data from a Port ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define read-uint8  (lambda (port) (read-string port 1)))
(define read-uint16 (lambda (port) (read-string port 2)))
(define read-uint32 (lambda (port) (read-string port 4)))
