;;; connection-requests.scm


;;;;;;;;;;;;;;;;;;;;;;;;
;;; ConnectionSetup  ;;;
;;; make-connection  ;;;
;;; close-connection ;;;
;;;;;;;;;;;;;;;;;;;;;;;;

;;;; Contact with the Xserver and get information about it.


;;; (in-path? filename)  => full-name or #f
;;; Is the file in our $PATH ?

;;; scshy.

(define in-path
  (lambda (fname)
    (let loop ((path exec-path-list))
      (if (null? path)
	  #f
	  (let ((full-name (path-list->file-name
			    (append (split-file-name (car path))
				    (list fname)))))
	    (if (file-executable? full-name)
		full-name
		(loop (cdr path))))))))

;;; scshy.

(define exists-and-executable?
  (lambda (filename)
    (and (file-exists? filename)
	 (file-executable? filename))))

;;; scshy.

(define get-hostname
  (lambda ()
    (let ((hostname-prog (in-path "hostname")))
      (if (exists-and-executable? hostname-prog)
	  (car (run/strings (,hostname-prog)))
	  (error "cannot determine hostname with ~a" hostname-prog)))))


;;; (parse-display-env)  => (values hostname display-number screen-number)
;;; figure out the hostname, display-number, and screen-number
;;; from the DISPLAY environment variable

;;; scshy.

(define parse-display-env
  (lambda ()
    (let ((display (getenv "DISPLAY"))
	  (display-regex "(.*):([0-9]+)[.]([0-9]+)"))
      (if (not display)
	  (error "can't read DISPLAY environment variable")
	  (let ((display-match (string-match display-regex display)))
	    (if (not display-match)
		(error "DISPLAY environment variable improperly formatted")
		(values (match:substring display-match 1)
			(match:substring display-match 2)
			(match:substring display-match 3))))))))


;;; for authorization code

(define hex-string->binary-data
  (lambda (str)
    (let* ((len     (string-length str))
           (new-len (quotient len 2))
           (new-str (make-string new-len)))
      (let loop ((i 0) (k 0))
        (if (< k new-len)
            (begin
	      ;; Umm, how about (substring str i (+ i 2))!
              (let* ((hex-str (string (string-ref str i)
				      (string-ref str (+ i 1))))
                     (hex-num (string->number hex-str 16))
                     (data    (string-ref (number->card8 hex-num) 0)))
                (string-set! new-str k data))
	      (loop (+ i 2) (+ k 1)))
	    new-str)))))


;;; (x-authorization hostname screen-number)  => (values protocol auth-code)
;;; Pull the X authorization information from .Xauthority (using xauth)
;;; A server doesn't necessarily require that that a machine use xauthority,
;;; so if we can't figure out the xauthority we just return something in the
;;; very slight possiblity that there's no security at all.

;;; scshy.

(define x-authorization
  (let ((xauthority-file ".Xauthority")
	(xauth-prog      "xauth"))
    (lambda (hostname screen)
      (if (let ((xauth (in-path xauth-prog)))
	    (not (and (file-readable? (home-file xauthority-file))
		      (exists-and-executable? xauth))))
	  (begin (format #t "~a not found; trying dummy values~%" xauthority-file)
		 (values "" ""))
	  (let ((screen (number->string screen)))
	    (let* ((machine (if #f  ; fix me -- a portable way?
				(string-append hostname "/unix" ":" screen)
				(string-append hostname ":" screen)))
		   (xauth-listing (let ((str (run/string (xauth list ,machine))))
				    (if (string=? str "")
					(begin
					  (format #t
						  "host (~a) not in ~a; trying dummy values~%"
						  machine xauthority-file)
					  (values "" ""))
					(let ((list ((infix-splitter) str)))
					  (if (not (>= (length list) 3))
					      (error "problem parsing xauth information")
					      list)))))
		   (protocol  (cadr xauth-listing))
		   (auth-code (hex-string->binary-data (caddr xauth-listing))))
	      (values protocol auth-code)))))))


;;; (determine-hostname-display-screen)
;;;   => (values hostname display-number screen-number)
;;; performs sanity checks on results from parse-display-env

(define determine-hostname-display-screen
  (lambda ()
    (bind ((machine-name display-number screen-number (parse-display-env)))
      (values (get-hostname)
	      (string->number display-number)
	      (string->number screen-number)
	      ))))


;;;; where it happens

(define endian 'little-endian)  ; How to determine this?
(define x-protocol-major 11)
(define x-protocol-minor  0)


;;; (connection-result xserver)
;;; read a byte from the xserver and decide whether it succeeded or failed.

(define connection-result
  (lambda (xserver)
    (case (read-card8 xserver)
      ((0) 'refused)
      ((1) 'success)
      ((2) 'more-auth)
      (else (error "impossibility in connection-result")))))


;;; (make-connection) => display
;;; Note: this doesn't lock send and reply.
;;; Nobody's trying to access a socket we haven't opened yet.

(define make-connection
  (lambda ()
    (bind ((hostname display-number screen-number (determine-hostname-display-screen))
	   (auth-protocol auth-code (x-authorization hostname display-number)))
      (let* ((xserver-port (+ 6000 display-number))
	     (xserver      (open-xserver-connection hostname xserver-port)))
	(if (not (xserver? xserver))
	    (error "Failed to open connection to xserver"
		   hostname display-number screen-number)
	    (bind ((protocol-n protocol-p padded-protocol (pad-data auth-protocol))
		   (code-n     code-p     padded-code     (pad-data auth-code)))
	      (let ((endian-ness (if (eq? endian 'big-endian) "B" "l")))
		(let ((str (make-string 12)))
		  (fill-string! str 0 endian-ness)
		  (number->card8/fill!  str  1 0)  ; unused
		  (number->card16/fill! str  2 x-protocol-major)
		  (number->card16/fill! str  4 x-protocol-minor)
		  (number->card16/fill! str  6 protocol-n)
		  (number->card16/fill! str  8 code-n)
		  (number->card16/fill! str 10 0)  ; unused
		  (send xserver str padded-protocol padded-code)
		  (case (connection-result xserver)
		    ((refused)   (error "Connection refused by Xserver"))
		    ((success)   (read-connection xserver))
		    ((more-auth) (error "Further authorization required")))))))))))


;;; (close-connection xserver)  => unspecified

(define close-connection
  (lambda (connection)
    (empty-event-queue)
    (close-xserver-connection
     (connection:xserver connection))))
