;;; set.scm
;;; Mike Wilson, March 17, 1999

;;; This has a bit-vector tinge to it.

;;; (make-set-type <set-size> <bit-name>*)  =>  <set-type>
;;;     <set-size> = <positive-integer>  ; such that <positive-integer> is a multiple of 8
;;;     <bit-name> = 0  ; always zero
;;;                = 1  ; always one
;;;                = <symbol>  ; such that <symbol> is the name of this bit
;;;                  ;; the rightmost bit is in bit-position zero.
;;;
;;; (set-type? <thing>)  =>  <boolean>

;;; (make-set <set-type> <initial-numeric-value>)  =>  <set>
;;;     <initial-numeric-value> = <positive-integer> such that it is < (expt 2 it)
;;;
;;; (set? <thing>)  =>  <boolean>

;;; (set-bit <set> <bit-name>*)  =>  unspecfied
;;;
;;; (clear-bit <set> <bit-name>*)  =>  unspecified
;;;
;;; (bit-set? <set> <bit>)  =>  <boolean>
;;;     <bit> = <symbol>  ; such that <symbol> is the name of a named bit
;;;     <bit> = <positive-integer>  ; such that <positive-integer> is <= <set-size>
;;;
;;; (bit-clear? <set> <bit>)  =>  <boolean>
;;;     as in bit-set?
;;;
;;; (set->symbols <set>)  =>  (<bit-name>*)
;;;     i.e. a bit-name for each named, set bit
;;;
;;; (set->number <set>)  =>  <positive-integer>


;;; Implementation
;;;
;;; <set-type> = #(set-type <set-size> <set-vector>)
;;; <set-size> = <integer>
;;; <set-vector> = #(<bit-name>*)
;;; <bit-name> = 0 | 1 | <symbol>
;;;
;;; <set> = #(set <set-type> <set-value>)
;;; <set-value> = <integer>

(define make-set-type-rep
  (lambda (set-size set-vector)
    (vector 'set-type set-size set-vector)))


;;; SET-TYPE?

(define set-type?
  (lambda (set-type)
    (and (vector? set-type)
	 (= (vector-length set-type) 3)
	 (eq? (vector-ref set-type 0) 'set-type))))

(define set-type-rep:size
  (lambda (set-type)
    (if (not (set-type? set-type))
	(error "expected set-type" set-type)
	(vector-ref set-type 1))))

(define set-type-rep:vector
  (lambda (set-type)
    (if (not (set-type? set-type))
	(error "expected set-type" set-type)
	(vector-ref set-type 2))))


(define make-set-rep
  (lambda (set-type initial-value)
    (vector 'set set-type initial-value)))


;;; SET?

(define set?
  (lambda (set)
    (and (vector? set)
	 (= (vector-length set) 3)
	 (eq? (vector-ref set 0) 'set))))

(define set-rep:type
  (lambda (set)
    (if (not (set? set))
	(error "expected a set" set)
	(vector-ref set 1))))

(define set-rep:value
  (lambda (set)
    (if (not (set? set))
	(error "expected a set" set)
	(vector-ref set 2))))

(define set-rep:value-set!
  (lambda (set new-value)
    (if (not (set? set))
	(error "expected a set" set)
	(vector-set! set 2 new-value))))

(define set-rep:size
  (lambda (set)
    (if (not (set? set))
	(error "expected a set" set)
	(let ((type (set-rep:type set)))
	  (set-type-rep:size type)))))

(define set-rep:vector
  (lambda (set)
    (if (not (set? set))
	(error "expected a set" set)
	(let ((type (set-rep:type set)))
	  (set-type-rep:vector type)))))


;;; MAKE-SET-TYPE

(define make-set-type
  (lambda (size . bit-names)
    (let ((number-of-bits (length bit-names)))
      (cond ((not (= number-of-bits size))
	     (error "number of bit-names not equal to given size" size bit-names))
	    ((not (zero? (modulo size 8)))
	     (error "set-size must be multiple of 8" size))
	    (else
	     (let ((set-vector (make-vector size 0)))
	       (let loop ((bit-position (- size 1)) (bit-names bit-names))
		 (if (null? bit-names)
		     (make-set-type-rep size set-vector)
		     (let ((bit-name (car bit-names)))
		       (cond ((symbol? bit-name)
			      (vector-set! set-vector bit-position bit-name))
			     ((not (integer? bit-name))
			      (error "unacceptable bit-name" bit-name))
			     ((= bit-name 0) (vector-set! set-vector bit-position 0))
			     ((= bit-name 1) (vector-set! set-vector bit-position 1))
			     (else (error "integer-named bit out of range" bit-name)))
		       (loop (- bit-position 1) (cdr bit-names)))))))))))


;;; MAKE-SET

(define make-set
  (lambda (set-type initial-value)
    (let* ((size    (set-type-rep:size set-type))
	   (largest (expt 2 size)))
      (if (< initial-value largest)
	  (make-set-rep set-type initial-value)
	  (error "initial-value larger than is acceptable for set-type" initial-value)))))


(define constant-bit?
  (lambda (set bit-name)
    (cond ((symbol? bit-name) #f)
	  ((not (integer? bit-name))
	   (error "bad bit-name" bit-name))
	  (else
	   (let ((vec  (set-rep:vector set))
		 (size (set-rep:size set)))
	     (if (< -1 bit-name size)
		 (symbol? (vector-ref vec bit-name))
		 (error "integer-named bit out of range" bit-name)))))))

;;; (define set-bit
;;;   (lambda (set . bit-names)
;;; 	(if (not (set? set))
;;; 	    (error "expected set" set)
;;; 	    (let ((set-type (set-rep:type set))
;;; 		  (type-vec (set-type-rep:vector set-type))
;;; 		  (vec-size (vector-length type-vec)))
;;; 	      (for-each (lambda (bit-name)
;;; 			  (if (integer? bit-name)
;;; 			      (if (or (negative? bit-name)
;;; 				      (> bit-name vec-size))
;;; 				  (error "integer-named bit out of range" bit-name)
;;; 				  (if (constant-bit? set bit-name)
;;; 				      (error "attempt to change constant bit" bit-name)
;;; 				      (set-rep:value-set!
;;; 				       set (bitwise-and value (expt 2 bit-name)))))
;;; 			      (let loop ((i 0))
;;; 				;; bit-position 0 is at vector position 0
;;; 				(cond ((> i vec-size)
;;; 				       (error "unknown bit-name" bit-name))
;;; 				      ((eq? (vector-ref type-vec i) bit-name)
;;; 				       (set-rep:value-set! set (bitwise-and value (expt 2 i))))
;;; 				      (else (loop (+ i 1)))))))
;;; 			bit-names)))))

(define name->position
  (lambda (set-type bit-name)
    (let* ((vec  (set-type-rep:vector set-type))
	   (size (vector-length vec)))
      (if (integer? bit-name)
	  (if (< -1 bit-name size)
	      bit-name
	      (error "integer-named bit out of range" bit-name))
	  (let loop ((i 0))
	    (cond ((>= i size) (error "illegal bit-name" bit-name))
		  ((eq? (vector-ref vec i) bit-name) i)
		  (else (loop (+ i 1)))))))))

(define zap-bit
  (lambda (set operation bit-names)
    (let ((set-type (set-rep:type set))
	  (value    (set-rep:value set)))
      (for-each (lambda (bit-name)
		  (if (constant-bit? set bit-name)
		      (error "attempt to change constant bit" bit-name)
		      (let ((bit-position (name->position set-type bit-name)))
			(set-rep:value-set! set (operation value (expt 2 bit-position))))))
		bit-names))))


;;; SET-BIT

(define set-bit
  (lambda (set . bit-names)
    (zap-bit set bitwise-ior bit-names)))


;;; CLEAR-BIT

(define clear-bit
  (lambda (set . bit-names)
    (zap-bit set bitwise-xor bit-names)))


;;; BIT-SET?

(define bit-set?
  (lambda (set bit-name)
    (let* ((set-type (set-rep:type set))
	   (value    (set-rep:value set))
	   (bit-position (name->position set-type bit-name)))
      (not (zero? (bitwise-and value (expt 2 bit-position)))))))


;;; BIT-CLEAR?

(define bit-clear?
  (lambda (set bit-name)
    (not (bit-set? set bit-name))))

(define set-member-count
  (lambda (set)
    (let ((size (set-rep:size set))
	  (vec  (set-rep:vector set)))
      (let loop ((bit 0) (count 0))
	(cond ((>= bit size) count)
	      ((bit-set? (vector-ref vec bit))
	       (loop (+ bit 1) (+ count 1)))
	      (else (loop (+ bit 1) count)))))))


;;; SET->SYMBOLS

(define set->symbols
  (lambda (set)
    (let ((vec   (set-rep:vector set))
	  (size  (set-rep:size set))
	  (value (set-rep:value set)))
      (let loop ((i (- size 1)) (symbols '() ))
	(if (negative? i)
	    symbols
	    (let ((bit-name (vector-ref vec i)))
	      (if (and (symbol? bit-name)
		       (not (zero? (bitwise-and value (expt 2 i)))))
		  (loop (- i 1) (cons bit-name symbols))
		  (loop (- i 1) symbols))))))))


;;; SET->NUMBER

(define set->number set-rep:value)
