;;; xprotocol-package.scm

;;;;;;;;;;;;;;;;;;
;;; Interfaces ;;;
;;;;;;;;;;;;;;;;;;

(define-interface non-portable-interface
  (export
   (define-record :syntax)
   null-char
   binary-string->printable-string
   bitwise-ior bitwise-and bitwise-not bitwise-xor arithmetic-shift
   format read-string char->ascii ascii->char
   make-lock with-lock
   string-match match:substring
   error
   ))

(define-interface tools-interface
  (export
   ;; syntax
   (bind :syntax) (when :syntax) (unless :syntax) (infinite-loop :syntax)

   ;; general tools
   append!  symbol-append
   first  second  third  fourth  fifth  sixth
   repeat  repeat-fn
   vector-for-each  identity

   ;; specific tools
   gather  pad-data  pad-length
   ))

(define-interface set-interface
  (export
   make-set-type set-type?
   make-set set?
   set-bit clear-bit bit-set?  bit-clear?
   set->symbols set->number
   ))

(define-interface xserver-interface
  (export
   open-xserver-connection  close-xserver-connection
   xserver-send  xserver-receive  xserver-peek  xserver?
   (with-xserver-lock :syntax)
   ))

(define-interface binary-data-interface
  (export
   number->uint8  number->uint8/fill!
   number->uint16 number->uint16/fill!
   number->uint32 number->uint32/fill!

   number->sint8  number->sint8/fill!
   number->sint16 number->sint16/fill!
   number->sint32 number->sint32/fill!

   uint8->number  uint16->number  uint32->number
   sint8->number  sint16->number  sint32->number
   read-uint8     read-uint16     read-uint32
 ))

(define-interface basic-conversions-interface
  (export
   number->card8   number->card8/fill!
   number->card16  number->card16/fill!
   number->card32  number->card32/fill!

   number->int8    number->int8/fill!
   number->int16   number->int16/fill!
   number->int32   number->int32/fill!

   card8->number     card16->number    card32->number
   int8->number      int16->number     int32->number
   string8->string   string->string8
   string16->string  string->string16
   ))

(define-interface basic-rw-interface
  (export
   read-int8     read-int16   read-int32
   read-card8    read-card16  read-card32
   read-string8
   send peek
   ))

(define-interface basic-datatypes-interface
  (compound-interface
   basic-rw-interface
   (export
    send  ignore  fill-string!

    number->card8  number->card16  number->card32
    number->int8   number->int16   number->int32

    number->card8/fill!  number->card16/fill!  number->card32/fill!
    number->int8/fill!   number->int16/fill!   number->int32/fill!

    number->atom       number->bit-gravity  number->bitmask
    number->button     number->colormap     number->cursor    number->depth
    number->drawable   number->font         number->fontable  number->gcontext
    number->keycode    number->keysym       number->pixmap
    number->timestamp  number->visual-id    number->window

    number->atom/fill!       number->bit-gravity/fill!
    number->bitmask/fill!    number->button/fill!     number->colormap/fill!
    number->cursor/fill!     number->depth/fill!      number->drawable/fill!
    number->font/fill!       number->fontable/fill!   number->gcontext/fill!
    number->keycode/fill!    number->keysym/fill!     number->pixmap/fill!
    number->timestamp/fill!  number->visual-id/fill!  number->window/fill!

    boolean->bool  boolean->bool/fill!
    symbol->bit-gravity  symbol->win-gravity

    symbols->set-of-event/fill!          read-set-of-event
    symbols->set-of-pointer-event/fill!  read-set-of-pointer-event
    symbols->set-of-device-event/fill!   read-set-of-device-event

    point-structures->points  arc-structures->arcs  rectangle-structures->rectangles
    internet-address->byte-sequence

    read-atom          read-bit-gravity  read-bitmask  read-bool       read-button
    read-colormap      read-cursor       read-depth    read-drawable   read-font
    read-fontable      read-gcontext     read-keycode  read-keysym     read-pixmap
    read-timestamp     read-visual-id    read-window   read-win-gravity
    read-list-of-bytes read-keybutmask   read-keymask

    symbol->value     value->symbol
    string8->string   string->string8
    string16->string  string->string16
    )))

(define-interface event-interface
  (export
   display-event
   button-press-event?  enter-notify-event?  expose-event?
   graphics-exposure-event?  key-press-event?  leave-notify-event?
   key-press-event:detail
   ))

(define-interface reply-interface
  (export
   send+read-reply

   read-alloc-color-reply           display-alloc-color-reply
   read-alloc-color-cells-reply     display-alloc-color-cells-reply
   read-alloc-color-planes-reply    display-alloc-color-planes-reply
   read-alloc-named-color-reply     display-alloc-named-color-reply
   read-get-screen-saver-reply      display-get-screen-saver-reply
   read-get-keyboard-mapping-reply  display-get-keyboard-mapping-reply
   read-list-extensions-reply       display-list-extensions-reply
   read-query-font-reply            display-query-font-reply
   read-query-text-extents-reply    display-query-text-extents-reply

   alloc-color-reply:pixel
   alloc-color-reply:red
   alloc-color-reply:green
   alloc-color-reply:blue

   alloc-named-color-reply:pixel

   query-text-extents-reply:overall-width
   query-text-extents-reply:overall-ascent
   query-text-extents-reply:overall-descent

   get-image-reply:data

   get-keyboard-mapping-reply:sequence-number
   get-keyboard-mapping-reply:keysyms-per-keycode
   get-keyboard-mapping-reply:keysyms

   get-screen-saver-reply:timeout
   get-screen-saver-reply:interval
   get-screen-saver-reply:prefer-blanking
   get-screen-saver-reply:allow-exposures
   ))

(define-interface xserver-read-interface
  (compound-interface
   event-interface
   reply-interface
   (export empty-event-queue read-event)
   ))

(define-interface structure-datatypes-interface
  (export
   new-id

   display-connection
   connection:xserver
   connection-xserver
   connection:protocol-major-version
   connection:protocol-minor-version
   connection:release-number
   connection:resource-id-base
   connection:resource-id-mask
   connection:motion-buffer-size
   connection:maximum-request-length
   connection:image-byte-order
   connection:bitmap-format-bit-order
   connection:bitmap-format-scanline-unit
   connection:bitmap-format-scanline-pad
   connection:min-keycode
   connection:max-keycode
   connection:vendor
   connection:pixmap-formats-list
   connection:root-screens-vector
   read-connection

   screen:root
   screen:default-colormap
   screen:white-pixel
   screen:black-pixel
   screen:current-input-masks
   screen:width-in-pixels
   screen:height-in-pixels
   screen:width-in-millimeters
   screen:height-in-millimeters
   screen:min-installed-maps
   screen:max-installed-maps
   screen:root-visual
   screen:backing-stores
   screen:save-unders
   screen:root-depth
   screen:depths-list

   string->string16  string16->string
   ))

(define-interface requests-interface
  (compound-interface
   event-interface
   reply-interface
   (export
    make-connection
    close-connection

    alloc-color alloc-color-cells alloc-color-planes alloc-named-color
    bell
    change-gc clear-area copy-area create-gc create-pixmap create-window
    get-image get-keyboard-mapping get-screen-saver
    image-text8
    list-extensions
    map-window
    no-operation
    open-font
    poly-fill-arc poly-fill-rectangle poly-line poly-point put-image
    query-font query-text-extents
    set-screen-saver
    )))

(define-interface helpers-interface
  (export
   (with-connection :syntax)
   connection-screen
   default-colormap
   root-depth root-visual root-window
   white-pixel black-pixel
   read-event
   ))

(define-interface scheme-xp-interface
  (compound-interface
   structure-datatypes-interface
   requests-interface
   helpers-interface
   tools-interface
   ))

(define-interface test-interface
  (export
   connect-test

   alloc-color-test alloc-color-cells-test
   alloc-color-planes-test alloc-named-color-test
   bell-test
   create-gc-test create-window-test
   get-keyboard-mapping-test
   image-text8-test
   list-extensions-test
   map-window-test
   no-operation-test
   poly-fill-arc-test poly-fill-rectangle-test
   poly-line-test     poly-point-test
   query-font-test query-text-extents-test
   screen-saver-test
   ))

(define-interface toys-interface
  (export
   hello-world hello-world-w/font
   bounce bounce-db bounce-th
   color-change
   mover
   shade
   ))

(define-interface sleep-interface
  (export
   unix-sleep
   thread-sleep))

;;;;;;;;;;;;;;;;
;;; Packages ;;;
;;;;;;;;;;;;;;;;

(define-structure non-portable non-portable-interface
  (open scheme scsh threads defrec-package)
  (files scsh-specific))

(define-structure tools tools-interface
  (open scheme non-portable)
  (files tools))

(define-structure set set-interface
  (open scheme non-portable)  ; bitwise
  (files set))

(define-structure binary-data binary-data-interface
  (open scheme non-portable tools)
  (files binary-data))

(define-structure basic-conversions basic-conversions-interface
  (open scheme non-portable tools binary-data)
  (files basic-conversions))

(define-structure xserver xserver-interface
  (open scheme non-portable scsh tools basic-conversions threads)
  (files xserver))

(define-structure basic-rw basic-rw-interface
  (open scheme tools basic-conversions xserver)
  (files basic-rw))

(define-structure basic-datatypes basic-datatypes-interface
  (open scheme non-portable set tools basic-conversions basic-rw)
  (files basic-datatypes))

(define-structure xserver-read xserver-read-interface
  (open scheme non-portable tools xserver basic-datatypes)
  (files replies events xserver-read))

(define-structure structure-datatypes structure-datatypes-interface
  (open scheme non-portable tools basic-datatypes)
  (files structure-datatypes))

(define-structure requests requests-interface
  ;; will take a little work to squeeze non-portabilities out of connection-requests
  (open scheme non-portable scsh tools set xserver basic-datatypes
	structure-datatypes xserver-read)
  (files connection-requests create-gc-request create-window-request requests))

(define-structure helpers helpers-interface
  (open scheme non-portable structure-datatypes requests xserver-read)
  (files helpers))

(define-structure sleep sleep-interface
  (open scheme structure-refs)
  (access scsh threads)
  (begin
    (define unix-sleep (structure-ref scsh sleep))
    (define thread-sleep (structure-ref threads sleep))))

(define-structure scheme-xp scheme-xp-interface
  (open scheme tools structure-datatypes requests helpers))

(define-structure test test-interface
  (open scheme scheme-xp formats sleep)
  (files test))

(define-structure toys toys-interface
  (open scheme scheme-xp formats threads)
  (files toys))

(define-structure frog (export froggy)
  (open scheme scheme-xp scsh defrec-package tools threads sleep fluids random)
  (files frog))
