;;; xserver-read.scm

;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Build an Event Queue ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; event queue
(define *events* '())
(define *event-lock* (make-lock))

(define dequeue-event
  (lambda (xserver)
    (with-lock *event-lock*
      (lambda ()
	(let ((event (car *events*)))
	  (set! *events* (cdr *events*))
	  event)))))

(define enqueue-event
  (lambda (event)
    (with-lock *event-lock*
      (lambda ()
	(set! *events* (append! *events* (list event)))
	;;(format #t "The Event Queue: ~s~%" *events*)
	))))

(define empty-event-queue
  (lambda ()
    (with-lock *event-lock*
      (lambda ()
	(set! *events* '() )))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Sit and Read from the Xserver ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define read-error
  (let ((throw-error (lambda (code code-description
			      sequence-number
			      extra-value extra-value-description
			      minor major)
		       (error "Xserver Error"
			      (format #f "~a (~a)" code-description code)
			      (format #f "sequence number: ~a" sequence-number)
			      (format #f "~a: ~a" extra-value-description extra-value)
			      (format #f "Minor Opcode: ~a" minor)
			      (format #f "Major Opcode: ~a" major)))))
    (lambda (xserver)
      (let* ((error-indicator (read-card8 xserver))
	     (code            (read-card8 xserver)))
	(bind ((code-description extra-value-description
		(case code
		  ((1) 	(values "Request Error"        "unused"))
		  ((2) 	(values "Value Error"          "bad value"))
		  ((3) 	(values "Window Error"         "bad resource id"))
		  ((4) 	(values "Pixmap Error"         "bad resource id"))
		  ((5) 	(values "Atom Error"           "bad atom id"))
		  ((6) 	(values "Cursor Error"         "bad resource id"))
		  ((7) 	(values "Font Error"           "bad rosource id"))
		  ((8) 	(values "Match Error"          "unused"))
		  ((9) 	(values "Drawable Error"       "bad resource id"))
		  ((10) (values "Access Error"         "unused"))
		  ((11) (values "Alloc Error"          "unused"))
		  ((12) (values "Colormap Error"       "bad resource id"))
		  ((13) (values "GContext Error"       "bad resource id"))
		  ((14) (values "IDChoice Error"       "bad resource id"))
		  ((15) (values "Name Error"           "unused"))
		  ((16) (values "Length Error"         "unused"))
		  ((17) (values "Implementation Error" "unused"))
		  (else (error "Unrecognized Error Code" code)))))
	      (let* ((sequence-number (read-card16 xserver))
		     (extra-value     (read-card32 xserver))
		     (minor-opcode    (read-card16 xserver))
		     (major-opcode    (read-card8 xserver))
		     (unused          (ignore 21 xserver)))
		(throw-error code code-description sequence-number
			     extra-value extra-value-description
			     minor-opcode major-opcode)))))))


(define read-event
  (lambda (xserver)
    (when (null? *events*)
      (clear-socket xserver))
    (dequeue-event xserver)))


;;; The use of args here is disturbing.  We have it so we can pass
;;; width*height to get-image so we can get clean data back.  It's
;;; either that or chop off padding from a returned string higher up.
;;; This approach, at least, is more general and may come up again.

(define read-reply
  (lambda (xserver type-expected . args)
    (if (clear-socket xserver)
	(case type-expected
	  ((47)  (read-query-font-reply xserver))
	  ((48)  (read-query-text-extents-reply xserver))
	  ((73)  (read-get-image-reply xserver (car args)))
	  ((84)  (read-alloc-color-reply xserver))
	  ((85)  (read-alloc-named-color-reply xserver))
	  ((86)  (read-alloc-color-cells-reply xserver))
	  ((87)  (read-alloc-color-planes-reply xserver))
	  ((99)  (read-list-extensions-reply xserver))
	  ((101) (read-get-keyboard-mapping-reply xserver))
	  ((108) (read-get-screen-saver-reply xserver))
	  (else  (error "Unrecognized reply code" type-expected)))
	(read-reply xserver type-expected))))


;;; clear-socket reads events out of the socket and queues them.
;;; We're either left with an empty queue or with a reply on top.
;;; Since we haven't added threadedness yet, we can't let this
;;; block or parts stop.
;;; For now, if we return #t, there's a reply waiting.
;;; #f if we don't know.

(define clear-socket
  (lambda (xserver)
    (let ((type (xserver-peek xserver)))
      (case type
	((0) (read-error xserver))	; error
	((1) #t)			; reply, leave it
	(else (begin (enqueue-event (redirect-event xserver type)) ; event
		     #f))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Atomic send and read ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; (send+read-reply <xserver> <package> <opcode>)
;;; (send+read-reply <xserver> <package> <opcode> <args>*)
;;; <package> = <string>
;;;           = (<string>*)
;;; <args> is usually omitted, but expected by a couple opcodes.

(define send+read-reply
  (lambda (xserver package-to-send opcode-num . opt)
    (with-xserver-lock
     (lambda ()
       (if (string? package-to-send)
	   (send xserver package-to-send)
	   (apply send (cons xserver package-to-send)))
       (apply read-reply (cons xserver (cons opcode-num opt)))))))
