;;; xserver.scm

;;;;;;;;;;;;;;;;;;;;;;;;;
;;; XServer Structure ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;

;;; in and out are relative to the server (or the socket).

(define *debug* #f)
(define debug (lambda () (set! *debug* (not *debug*))))

(define-record xserver
  socket
  in
  out
  open?
  )

(define *send-lock* (make-lock))
(define *read-lock* (make-lock))

(define-syntax with-xserver-lock
  (syntax-rules ()
    ((with-xserver-lock thunk)
     (with-lock *send-lock* (lambda () (with-lock *read-lock* thunk))))))

(define open-xserver-connection
  (lambda (hostname port-number)
    (with-lock *send-lock*
     (lambda ()
       (let* ((socket (socket-connect protocol-family/internet
				      socket-type/stream
				      hostname
				      port-number))
	      (in  (socket:inport socket))
	      (out (socket:outport socket)))
	 (set-port-buffering in bufpol/none)
	 (set-port-buffering out bufpol/none)
	 (make-xserver socket in out #t))))))


(define dead-connection-error
  (lambda ()
    (error "Connection to Xserver killed")))


(define xserver-send
  (lambda (xserver . data)
    (with-lock *send-lock*
     (lambda ()
       (for-each (lambda (str)
		   (write-string str (xserver:out xserver)))
		 data)
       (when *debug*
	 (format #t "xserver-send wrote ~a~%"
		 (binary-string->printable-string (apply string-append data))))
       ))))


(define xserver-receive
  (lambda (xserver length)
    (with-lock *read-lock*
      (lambda ()
	(let ((str (read-string length (xserver:in xserver))))
	  (when *debug* (format #t "xserver-receive got:  ~a~%"
				(binary-string->printable-string str)))
	  (if (eof-object? str)
	      (begin (close-xserver-connection xserver)
		     (dead-connection-error))
	      str))))))


(define xserver-peek
  (lambda (xserver)
    (with-lock *read-lock*
      (lambda ()
	(let ((response (peek-char (xserver:in xserver)))) ; blocks
	  (if (eof-object? response)
	      (begin (close-xserver-connection xserver)
		     (dead-connection-error))
	      (card8->number (string response))))))))

(define close-xserver-connection
  (lambda (xserver)
    (with-lock *send-lock*
      (lambda ()
	(set-xserver:open? xserver #f)
	(close-socket (xserver:socket xserver))))))
